@@ -5,13 +5,13 @@ The main core of EasyTorch
 import math as _math
 import os as _os
 from collections import OrderedDict as _ODict
-from typing import List as _List
+from typing import List as _List, Union as _Union
 
 import torch as _torch
 from torch.utils.data import Dataset as _Dataset
 
 import easytorch.utils as _etutils
-from easytorch.config.status import *
+from easytorch.config.state import *
 from easytorch.metrics import metrics as _base_metrics
 from easytorch.utils.logger import *
 from easytorch.utils.tensorutils import initialize_weights as _init_weights
@@ -228,19 +228,34 @@ class ETTrainer:
     def evaluation(self,
                    epoch=1,
                    mode='eval',
-                   dataset_list=None,
-                   save_pred=False) -> dict:
+                   dataset=_Union[_List[_Dataset], _Dataset],
+                   save_pred=False,
+                   distributed: bool = False,
+                   use_unpadded_sampler: bool = False) -> dict:
 
         for k in self.nn:
             self.nn[k].eval()
 
         eval_avg, eval_metrics = self.new_averages(), self.new_metrics()
 
-        if dataset_list is None:
+        if dataset is None:
             return {'averages': eval_avg, 'metrics': eval_metrics}
 
         info(f'{mode} ...', self.args['verbose'])
-        loaders = [self.data_handle.get_loader(handle_key=mode, shuffle=False, dataset=d) for d in dataset_list]
+        if not isinstance(dataset, list):
+            dataset = [dataset]
+
+        loaders = []
+        for d in dataset:
+            loaders.append(
+                self.data_handle.get_loader(
+                    handle_key=mode,
+                    shuffle=False, dataset=d,
+                    distributed=distributed,
+                    use_unpadded_sampler=use_unpadded_sampler
+                )
+            )
+
         with _torch.no_grad():
             for loader in loaders:
                 its = []
@@ -256,20 +271,20 @@ class ETTrainer:
                     if save_pred:
                         its.append(it)
 
-                    if self.args['verbose'] and len(dataset_list) <= 1 and lazy_debug(i, add=epoch):
+                    if self.args['verbose'] and len(dataset) <= 1 and lazy_debug(i, add=epoch):
                         info(
                             f" Itr:{i}/{len(loader)}, Averages:{it.get('averages').get()}, Metrics:{it.get('metrics').get()}")
 
                 eval_metrics.accumulate(metrics)
                 eval_avg.accumulate(avg)
 
-                if self.args['verbose'] and len(dataset_list) > 1:
+                if self.args['verbose'] and len(dataset) > 1:
                     info(f" {mode}, {avg.get()}, {metrics.get()}")
                 if save_pred:
                     self.save_predictions(loader.dataset, self._reduce_iteration(its))
 
         info(f"{self.cache['experiment_id']} {mode} Averages:{eval_avg.get()}, Metrics:{eval_metrics.get()}",
-                self.args['verbose'])
+             self.args['verbose'])
 
         return {'averages': eval_avg, 'metrics': eval_metrics}
 
@@ -354,7 +369,7 @@ class ETTrainer:
                 self.optimizer[optim].zero_grad()
         return it
 
-    def reduce_scores(self, accumulator: list) -> dict:
+    def reduce_scores(self, accumulator: list, distributed=False) -> dict:
         averages = self.new_averages()
         metrics = self.new_metrics()
         if all([a is None for a in accumulator]):
@@ -365,7 +380,7 @@ class ETTrainer:
             averages.accumulate(acc['averages'])
             metrics.accumulate(acc['metrics'])
 
-        if self.args['use_ddp']:
+        if distributed:
             avg_serial = _torch.tensor(averages.serialize()).to(self.device['gpu'])
             _dist.reduce(avg_serial, dst=MASTER_RANK, op=_dist.ReduceOp.SUM)
 
@@ -393,8 +408,11 @@ class ETTrainer:
                 f"Not best: {val_check['score']}, {self.cache['best_val_score']} in ep: {self.cache['best_val_epoch']}",
                 self.args['verbose'])
 
-    def validation(self, epoch, val_dataset_list: _List[_Dataset]) -> dict:
-        return self.evaluation(epoch=epoch, mode='validation', dataset_list=val_dataset_list)
+    def validation(self, epoch, dataset: _Union[_List[_Dataset], _Dataset]) -> dict:
+        return self.evaluation(epoch=epoch, mode='validation',
+                               dataset=dataset,
+                               distributed=self.args['use_ddp'],
+                               use_unpadded_sampler=True)
 
     def _global_debug(self, running_averages, running_metrics, **kw):
         """Update running accumulators."""
@@ -402,7 +420,7 @@ class ETTrainer:
         running_metrics.accumulate(kw.get('metrics'))
 
         """Reset iteration accumulator"""
-        N = kw['tot_iter']
+        N = kw['num_iters']
         i, e = kw['i'], kw['epoch']
 
         if lazy_debug(i, add=e) or i == N:
@@ -417,11 +435,16 @@ class ETTrainer:
 
             running_averages.reset(), running_metrics.reset()
 
-    def train(self, train_dataset: _Dataset, val_dataset_list: _List[_Dataset]) -> None:
+    def train(self, train_dataset: _Dataset, validation_dataset: _Union[_List[_Dataset], _Dataset]) -> None:
         info('Training ...', self.args['verbose'])
-        train_loader = self.data_handle.get_loader(handle_key='train', shuffle=True, dataset=train_dataset)
 
-        tot_iter = len(train_loader) // self.args['grad_accum_iters']
+        train_loader = self.data_handle.get_loader(
+            handle_key='train',
+            shuffle=True,
+            dataset=train_dataset,
+            distributed=self.args['use_ddp']
+        )
+
         for ep in range(1, self.args['epochs'] + 1):
             for k in self.nn:
                 self.nn[k].train()
@@ -438,31 +461,36 @@ class ETTrainer:
             if self.args.get('use_ddp'):
                 train_loader.sampler.set_epoch(ep)
 
+            num_iters = len(train_loader) // self.args['grad_accum_iters']
             for i, batch in enumerate(train_loader, 1):
                 its.append(self.training_iteration(i, batch))
                 """When end of iteration"""
                 if i % self.args['grad_accum_iters'] == 0:
                     it = self._reduce_iteration(its)
 
-                    its = []
-                    it['epoch'] = ep
-                    it['i0'], it['tot_iter'] = i, tot_iter
-                    it['i'] = i // self.args['grad_accum_iters']
-
                     """Update global accumulators"""
+                    its = []
+                    it['num_iters'] = num_iters
+                    it['i'] = i // self.args['grad_accum_iters']
                     epoch_avg.accumulate(it.get('averages'))
                     epoch_metrics.accumulate(it.get('metrics'))
 
                     if self.args['is_master']:
-                        self._global_debug(_avg, _metrics, **it)
-                    self._on_iteration_end(i=i, ep=ep, it=it)
+                        self._global_debug(_avg, _metrics, epoch=ep, **it)
+                    self._on_iteration_end(i=i, epoch=ep, it=it)
+
+            reduced_epoch = self.reduce_scores(
+                [{'averages': epoch_avg, 'metrics': epoch_metrics}],
+                distributed=self.args['use_ddp']
+            )
+            epoch_out = {'epoch': ep, 'training': reduced_epoch}
 
             """Validation step"""
-            reduced_epoch = self.reduce_scores([{'averages': epoch_avg, 'metrics': epoch_metrics}])
-            epoch_out = {'epoch': ep, 'training': reduced_epoch}
-            if val_dataset_list:
-                val_out = self.validation(ep, val_dataset_list)
-                epoch_out['validation'] = self.reduce_scores([val_out])
+            if validation_dataset is not None:
+                val_out = self.validation(ep, validation_dataset)
+                if self.args['use_ddp']:
+                    _dist.barrier()
+                epoch_out['validation'] = self.reduce_scores([val_out], distributed=self.args['use_ddp'])
 
             if self.args['is_master']:
                 self._global_epoch_end(**epoch_out)
